// simple temperture program

#define TERMO_PIN 9

const static char hv[] = "0123456789ABCDEF";

static void print_hex(const unsigned char *data, int len) {
   unsigned char i = 0;
   while (i < len) {
     Serial.print(hv[data[i] >> 4]);
     Serial.print(hv[data[i] & 15]);
     i++;
   }
}

//========================================= DS18S20 Temperature chip i/o

#if 0
/* pin 15 on Sanguino = PD7  */
#define W1PORT PORTD
#define W1DDR  DDRD
#define W1PIN  PIND
#define W1BIT  PORTD7
#endif

/* pin 8 on Arduino */
#define W1PORT PORTB
#define W1DDR  DDRB
#define W1PIN  PINB
#define W1BIT  PORTB0

/* enable/disable the use of the internal pullup.
   It also uses output driver for the idle state - that may enhance stability but increases power usage.
   Note that the specs ask for 5k pullup whereas ATmega has 10k, but our timings are relaxed
   enough that it should still work in most situations.
   */
#define INT_PULLUP 1

#define nop() __asm__("nop\n\t")
#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif 

void w1_setup() {
#if INT_PULLUP
  sbi(W1PORT, W1BIT); /* enable pullup */
#else
  cbi(W1PORT, W1BIT); /* disable pullup */
#endif
  cbi(W1DDR, W1BIT); /* init in idle (input) mode in either case - we'll be driving the bus on the reset pulse */
}

/* 0 on success */
static char w1_reset() {
  sbi(W1PORT, W1BIT); /* the bus should already be high, but just in case ... */
  sbi(W1DDR, W1BIT);
  /* let the device get a high pulse for charging */
  delayMicroseconds(10);
  /* pull 1W low for 480us */
  cbi(W1PORT, W1BIT);
  delayMicroseconds(480);
  sbi(W1PORT, W1BIT);
  delayMicroseconds(2); /* quick power-on in case the devices need it (this is not accordingh to specs but should help the recovery) */
  /* switch to input mode */
  cbi(W1DDR, W1BIT);
#if ! INT_PULLUP
  cbi(W1PORT, W1BIT); /* disable pull-up */
#endif
  /* wait until there is a presence pulse or at most 480us */
  unsigned long to = micros() + 480;
  char ps = 1;
  while ((ps = (W1PIN & _BV(W1BIT))) && micros() < to) ;
  if (!ps) { /* ok, pulse found */
    /* line is low,  wait for it to got back to high */
    to = micros() + 600; // extend the wait beyond 480 just in case some other device attempts to pull low ...
    while ((ps = (W1PIN & _BV(W1BIT))) == 0 && micros() < to) ;
    if (ps) {
      ps = 0; delayMicroseconds(15); /* just in case let the bus settle */
    } else ps = 1;
  }
  if (!ps) { /* power-up the bus if there are devices, otherwise we keep it unpowered */
    sbi(W1PORT, W1BIT); /* enable pull-up and put the bus back to driving mode */
    sbi(W1DDR, W1BIT);
  }
  return ps;
}

/* keeps the bus in powered state */
static void w1_write8(unsigned char val) {
  sbi(W1PORT, W1BIT); /* the bus should already be high, but just in case ... */
  sbi(W1DDR, W1BIT);
  for (char i = 0; i < 8; i++) {
    if (val & 1) {
      cbi(W1PORT, W1BIT); /* low */
      delayMicroseconds(4); /* this can be really anything from 4-8 */
      sbi(W1PORT, W1BIT); /* high */
      delayMicroseconds(65); /* the slot + recovery will be about 70us */  
    } else {
      cbi(W1PORT, W1BIT); /* low */
      delayMicroseconds(60); /* keep it low for the entire time of the slot */
      sbi(W1PORT, W1BIT); /* high */
      delayMicroseconds(10); /* recovery */
    }
    val >>= 1;
  }
}

static unsigned char w1_read8() {
  unsigned char val = 0;
  sbi(W1PORT, W1BIT);
  sbi(W1DDR, W1BIT); /* enable driver */
  for (char i = 0; i < 8; i++) {
      cbi(W1PORT, W1BIT); /* pull low */
      /* we need to wait 1us but delayMicroseconds gives us at best 4us .. we'll see */
      delayMicroseconds(2);
      cbi(W1DDR, W1BIT); /* back to input mode */
#if INT_PULLUP
      sbi(W1PORT, W1BIT); /* enable pull-up */
#endif
      /* the specs guarantee the value for 15us since we pulled the line low.
         For hard devices we could acually infer the case-specific timing from the ROM READ
         response, but it seems taking the edge case of 15us since *release* seems to work reasonably.
         Shorter times don't leave enough room for 1 bits to rise via pullup, longer times
         will sample after the slave has released the line. Fortunately both effects have usually
         the same rise time so the time could be adjusted according to the wiring/pullup characteristics.*/
      delayMicroseconds(17);
      val >>= 1;
      if (W1PIN & _BV(W1BIT)) val |= (unsigned char) 0x80;
      delayMicroseconds(43);
      sbi(W1PORT, W1BIT);
      sbi(W1DDR, W1BIT); /* enable output for recovery */
      delayMicroseconds(10);
  }
  return val;
}

/* this function puts the bus into idle state by de-powering the bus - i.e. releasing it to the pull-up */
static void w1_done() {
#if INT_PULLUP
  sbi(W1PORT, W1BIT);
#endif
  cbi(W1DDR, W1BIT); /* go to input more - releasing the bus */
#if ! INT_PULLUP
  cbi(W1PORT, W1BIT); /* remove internal pullup */
#endif
}

/* just a wrapper for read8 to read into a buffer */
static void w1_read(unsigned char *data, unsigned char len = 8)
{
  for (unsigned char i = 0; i < len; i++) data[i] = w1_read8();
}

/* CRC based on X^8 + X^5 + X^4 + 1 */
static unsigned char crc8(unsigned char data, unsigned char iv) {
  for (unsigned char i = 0; i < 8; i++) {
    iv = ((iv ^ data) & 1) ? ((iv >> 1) ^ 0x8C) : (iv >> 1);
    data >>= 1;
  }
  return iv;
}

/* same as above but will automatically feed an array with the initial value of 0 */
static unsigned char crc8(unsigned char *data, unsigned char len) {
  unsigned char iv = 0;
  for (unsigned char i = 0; i < len; i++) iv = crc8(data[i], iv);
  return iv;
}

/* converts 11-bit temperature from the DS18B20 tr into 1/100th of deg.C */
static int sp2temp(const unsigned char *data) {
  unsigned int iv = data[0] | (data[1] << 8), rv;
  char sign = (iv & 0x8000) ? 1 : 0;  // test most sig bit
  if (sign) iv = (~iv) + 1;
  rv = 6 * iv + iv / 4;
  return sign ? -rv : rv;
}

/* 1-Wire */
#define READ_ROM     0x33
#define SEARCH_ROM   0xF0
#define MATCH_ROM    0x55
#define SKIP_ROM     0xCC
/* DS18B20 */
#define ALARM_SEARCH 0xEC
#define CONVERT_T    0x44
#define WRITE_SP     0x4E /* TH, TL, CONF */
#define READ_SP      0xBE /* b0...b7 + CRC */

/* reset + match ROM + send command. if data is NULL then SKIP ROM is issued instead of MATCH ROM.
   if command is 0 then no command is sent.
   returns 1 is reset failed (no devices), 0 otherwise */
static int w1_talk_to(const unsigned char *addr, unsigned char cmd = 0) {
  if (w1_reset()) return 1;
  delayMicroseconds(60); // wait a bit after the presence pulse ... just in case ... 
  if (!addr)
    w1_write8(SKIP_ROM);
  else {
    w1_write8(MATCH_ROM);
    for (unsigned char i = 0; i < 8; i++) w1_write8(addr[i]);
  }
  if (cmd) w1_write8(cmd);
  return 0;
}



//========================================= SETUP

void setup() {
  pinMode(13, OUTPUT);
  pinMode(TERMO_PIN, INPUT);
  digitalWrite(TERMO_PIN, HIGH); /* enable pull-up */
  w1_setup();
  Serial.begin(9600);
  Serial.println("minitemp 1.0-0");
}

//========================================= MAIN

static char termo = LOW;

static unsigned char address[16 * 8] = {
    0x28, 0x08, 0x2F, 0x18, 0x02, 0x00, 0x00, 0x43, /* #1 */
    0x28, 0xA1, 0x31, 0x18, 0x02, 0x00, 0x00, 0x57, /* #3  R  basement boiler */
    0x28, 0xD4, 0x2B, 0x18, 0x02, 0x00, 0x00, 0xE5, /* #2  Y  living room */
    0x28, 0x16, 0x23, 0x18, 0x02, 0x00, 0x00, 0x2A, /* #4  G  outside */
    0x28, 0xEF, 0x5C, 0x18, 0x02, 0x00, 0x00, 0xCC, /* #5  B  basement, desk */
    0x28, 0x0F, 0xDF, 0xF6, 0x01, 0x00, 0x00, 0xD8, /* #6  RY (built-in boiler) */
    0x28, 0x66, 0xB2, 0x8F, 0x02, 0x00, 0x00, 0xAF, /* #8  P  */
    0x28, 0xD1, 0x8B, 0x8F, 0x02, 0x00, 0x00, 0x5A, /* #9  RP */
    0x28, 0xBE, 0x8E, 0x8F, 0x02, 0x00, 0x00, 0x6F, /* #10 RG */
0 };
static unsigned char addresses = 8 * 9;

static unsigned char tsp[9];

// this function is called while waiting
static void activeRun() {
  /* termo pin monitor */
  char cte = digitalRead(TERMO_PIN);
  if (cte != termo) {
    termo = cte;
    Serial.print("TS ");
    Serial.print(millis());
    Serial.println(cte ? " ON" : " OFF");
  }
}

static void activeDelay(unsigned int amount) {
  unsigned int st = millis(), et = st + amount;
  if (et > st) { /* easy part - no overflow */
    while (1) {
      activeRun();
      unsigned int ct = millis();
      if (ct > et || ct < st) return; /* the second condition is there just in case the et is on the edge of the overflow */
    }
  } else { /* overflow */
    while (1) {
      activeRun();
      unsigned int ct = millis();
      if (ct > et && ct < st) return; /* in the overflow case the time must fall between et and st */
    }
  }
  /* unreachable */
}

void loop() {
  digitalWrite(13, HIGH);
  /* tell all devices to measure temp */
  if (!w1_talk_to(0, CONVERT_T)) { /* bother only if there is anything on the wire */
    /* the current code guarantees the bus to be in powered mode so parasite power is fine */
    activeDelay(800); /* wait 800ms for conversion */
    unsigned char q = 0;
    while (q < addresses) {
      w1_talk_to(address + q, READ_SP);
      tsp[0] = w1_read8();
      if (tsp[0] != 0xFF) { /* bother continuing only if it's a valid response */
        w1_read(tsp + 1, 8);
        Serial.print('T');
        print_hex(address + q, 8);
        Serial.println(": ");
        if (crc8(tsp, 9) == 0) { /* check CRC */
          if (tsp[0] == 0x50 && tsp[1] == 0x05) /* power-up default - we choose to ignore it for now */
            Serial.println("power-up default, ignoring");
          else {  
            int temp = sp2temp(tsp);
            if (temp < 0) { temp = -temp; Serial.print('-'); }
            Serial.print(temp / 100);
            Serial.print('.');
            temp %= 100;
            if (temp < 10) {
              Serial.print('0');
              Serial.println(((char)temp) + '0');
            } else {
              Serial.println(temp);
            }
          }
        } else Serial.println("invalid CRC");
      }
      q += 8;
    }
  } else Serial.println("(nothing on 1-wire)\n");
  w1_done(); /* de-power the bus */
  digitalWrite(13, LOW);

  activeDelay(10000);
}
