/* routines for 1-Wire interface and DS18B20 temp. sensors
 *
 * (C)Copyright 2009,10 Simon Urbanek
 *
 * NOTE: requires W1PORT, W1DDR, W1PIN and W1BIT to be defined (see EOF for minimal example).
 *
 * Optional: INT_PULLUP if defined to 1 then internal pullup is used instead
 *                      of the external pullup.
 *
 *           NO_DS18B20 ommit code and definitions for DS18B20
 *
 * The code is C-safe, but requires delayMicroseconds() and micros().
 * The latter doesn't have to be precise, it is used to measure the presence pulse
 * so a simple counter implementation will do.
 *
 * License: GPL version 2
 */

#ifndef W1_H__
#define W1_H__

#include <avr/io.h>

/* 1-Wire interface functions */

static void w1_setup();  /* setup ports and directions - needed only once */

static char w1_reset();  /* reset -- returns 0 if there are devices, 1 if that are none */

static void w1_write8(unsigned char val);

static unsigned char w1_read8();

static void w1_done();   /* de-powers the bus */

/* combines reset + SKIP_ROM(addr=0)/MATCH_ROM + write command (cmd!=0); returns the result of reset */
#ifdef __cplusplus
static char w1_talk_to(const unsigned char *addr, unsigned char cmd = 0);
#else
static char w1_talk_to(const unsigned char *addr, unsigned char cmd);
#endif

/* 1-Wire utility functions */
static unsigned char crc8(unsigned char data, unsigned char iv);

#ifdef __cplusplus
static unsigned char crc8(unsigned char *data, unsigned char len);
#else
static unsigned char crc8_data(unsigned char *data, unsigned char len);
#endif
	
#ifndef NO_DS18B20
static int sp2temp(const unsigned char *data);
#endif

/*---- implementation ----*/

#ifndef nop
#define nop() __asm__("nop\n\t")
#endif
#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif 

static void w1_setup() {
#if INT_PULLUP
	sbi(W1PORT, W1BIT); /* enable pullup */
#else
	cbi(W1PORT, W1BIT); /* disable pullup */
#endif
	cbi(W1DDR, W1BIT); /* init in idle (input) mode in either case - we'll be driving the bus on the reset pulse */
}

/* 0 on success */
static char w1_reset() {
	sbi(W1PORT, W1BIT); /* the bus should already be high, but just in case ... */
	sbi(W1DDR, W1BIT);
	/* let the device get a high pulse for charging */
	delayMicroseconds(10);
	/* pull 1W low for 480us */
	cbi(W1PORT, W1BIT);
	delayMicroseconds(480);
	sbi(W1PORT, W1BIT);
	delayMicroseconds(2); /* quick power-on in case the devices need it (this is not accordingh to specs but should help the recovery) */
	/* switch to input mode */
	cbi(W1DDR, W1BIT);
#if ! INT_PULLUP
	cbi(W1PORT, W1BIT); /* disable pull-up */
#endif
	/* wait until there is a presence pulse or at most 480us */
	unsigned long to = micros() + 480;
	char ps = 1;
	while ((ps = (W1PIN & _BV(W1BIT))) && micros() < to) ;
	if (!ps) { /* ok, pulse found */
		/* line is low,  wait for it to got back to high */
		to = micros() + 600; // extend the wait beyond 480 just in case some other device attempts to pull low ...
		while ((ps = (W1PIN & _BV(W1BIT))) == 0 && micros() < to) ;
		if (ps) {
			ps = 0; delayMicroseconds(15); /* just in case let the bus settle */
		} else ps = 1;
	}
	if (!ps) { /* power-up the bus if there are devices, otherwise we keep it unpowered */
		sbi(W1PORT, W1BIT); /* enable pull-up and put the bus back to driving mode */
		sbi(W1DDR, W1BIT);
	}
	return ps;
}

/* keeps the bus in powered state */
static void w1_write8(unsigned char val) {
	sbi(W1PORT, W1BIT); /* the bus should already be high, but just in case ... */
	sbi(W1DDR, W1BIT);
	for (char i = 0; i < 8; i++) {
		if (val & 1) {
			cbi(W1PORT, W1BIT); /* low */
			delayMicroseconds(4); /* this can be really anything from 4-8 */
			sbi(W1PORT, W1BIT); /* high */
			delayMicroseconds(65); /* the slot + recovery will be about 70us */  
		} else {
			cbi(W1PORT, W1BIT); /* low */
			delayMicroseconds(60); /* keep it low for the entire time of the slot */
			sbi(W1PORT, W1BIT); /* high */
			delayMicroseconds(10); /* recovery */
		}
		val >>= 1;
	}
}

static unsigned char w1_read8() {
	unsigned char val = 0;
	sbi(W1PORT, W1BIT);
	sbi(W1DDR, W1BIT); /* enable driver */
	for (char i = 0; i < 8; i++) {
		cbi(W1PORT, W1BIT); /* pull low */
		/* we need to wait 1us but delayMicroseconds gives us at best 4us .. we'll see */
		delayMicroseconds(2);
		cbi(W1DDR, W1BIT); /* back to input mode */
#if INT_PULLUP
		sbi(W1PORT, W1BIT); /* enable pull-up */
#endif
		/* the specs guarantee the value for 15us since we pulled the line low.
		 For hard devices we could acually infer the case-specific timing from the ROM READ
		 response, but it seems taking the edge case of 15us since *release* seems to work reasonably.
		 Shorter times don't leave enough room for 1 bits to rise via pullup, longer times
		 will sample after the slave has released the line. Fortunately both effects have usually
		 the same rise time so the time could be adjusted according to the wiring/pullup characteristics.*/
		delayMicroseconds(17);
		val >>= 1;
		if (W1PIN & _BV(W1BIT)) val |= (unsigned char) 0x80;
		delayMicroseconds(43);
		sbi(W1PORT, W1BIT);
		sbi(W1DDR, W1BIT); /* enable output for recovery */
		delayMicroseconds(10);
	}
	return val;
}

/* this function puts the bus into idle state by de-powering the bus - i.e. releasing it to the pull-up */
static void w1_done() {
#if INT_PULLUP
	sbi(W1PORT, W1BIT);
#endif
	cbi(W1DDR, W1BIT); /* go to input more - releasing the bus */
#if ! INT_PULLUP
	cbi(W1PORT, W1BIT); /* remove internal pullup */
#endif
}

#ifdef __cplusplus
/* just a wrapper for read8 to read into a buffer */
static void w1_read(unsigned char *data, unsigned char len = 8)
{
	for (unsigned char i = 0; i < len; i++) data[i] = w1_read8();
}
#endif

/* CRC based on X^8 + X^5 + X^4 + 1 */
static unsigned char crc8(unsigned char data, unsigned char iv) {
	for (unsigned char i = 0; i < 8; i++) {
		iv = ((iv ^ data) & 1) ? ((iv >> 1) ^ 0x8C) : (iv >> 1);
		data >>= 1;
	}
	return iv;
}

/* same as above but will automatically feed an array with the initial value of 0 */
#ifdef __cplusplus
static unsigned char crc8(unsigned char *data, unsigned char len)
#else
static unsigned char crc8_data(unsigned char *data, unsigned char len)
#endif
{
	unsigned char iv = 0;
	for (unsigned char i = 0; i < len; i++) iv = crc8(data[i], iv);
	return iv;
}

/* 1-Wire */
#define READ_ROM     0x33
#define SEARCH_ROM   0xF0
#define MATCH_ROM    0x55
#define SKIP_ROM     0xCC

#ifndef NO_DS18B20
/* DS18B20 */
#define ALARM_SEARCH 0xEC
#define CONVERT_T    0x44
#define WRITE_SP     0x4E /* TH, TL, CONF */
#define READ_SP      0xBE /* b0...b7 + CRC */

/* converts 11-bit temperature from the DS18B20 tr into 1/100th of deg.C */
static int sp2temp(const unsigned char *data) {
	unsigned int iv = data[0] | (data[1] << 8), rv;
	char sign = (iv & 0x8000) ? 1 : 0;  // test most sig bit
	if (sign) iv = (~iv) + 1;
	rv = 6 * iv + iv / 4;
	return sign ? -rv : rv;
}
#endif

/* reset + match ROM + send command. if data is NULL then SKIP ROM is issued instead of MATCH ROM.
 if command is 0 then no command is sent.
 returns 1 is reset failed (no devices), 0 otherwise */
static char w1_talk_to(const unsigned char *addr, unsigned char cmd) {
	if (w1_reset()) return 1;
	delayMicroseconds(60); // wait a bit after the presence pulse ... just in case ... 
	if (!addr)
		w1_write8(SKIP_ROM);
	else {
		w1_write8(MATCH_ROM);
		for (unsigned char i = 0; i < 8; i++) w1_write8(addr[i]);
	}
	if (cmd) w1_write8(cmd);
	return 0;
}
	
#endif

/* minimal example:

#include <avr/io.h>

#define W1PORT PORTD
#define W1DDR  DDRD
#define W1PIN  PIND
#define W1BIT  PORTD7

#include <w1.h>

static char tsp[9];
 
void loop() {
  // tell all devices to convert
  if (!w1_talk_to(0, CONVERT_T)) { // bother only if there is anything on the wire
    w1_talk_to(address + q, READ_SP);
    tsp[0] = w1_read8();
    if (tsp[0] != 0xFF) { // bother continuing only if it's a valid response
      w1_read(tsp + 1, 8);
      int temp = sp2temp(tsp);
    }
  }
  delay(10000);
}

*/
